#------------------------------------------------------------------------------
#
#  Copyright (c) 2014-2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------
""" Dummy objects for `TypeRegistry` testing.
"""

import abc
import six


class A(object):
    pass


class B(A):
    pass


class C(B):
    pass


class D(object):
    pass


class Mixed(A, D):
    pass


@six.add_metaclass(abc.ABCMeta)
class Abstract(object):
    pass


class Concrete(object):
    pass


class ConcreteSubclass(Concrete):
    pass


for typ in (A, B, C, D, Mixed, Abstract, Concrete, ConcreteSubclass):
    typ.__module__ = 'dummies'


Abstract.register(Concrete)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2014-2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------

from traits.api import Any, Callable, Dict, Either, HasPrivateTraits, \
    Instance, List, Str, Undefined
from traitsui.editor_factory import EditorFactory
from traitsui.item import Item
from traitsui.qt4.editor import Editor

from .binder import Binder
from .binding import Binding
from .qt import QtGui
from .raw_widgets import ButtonGroup


class TraitsUI(Binder):
    """ Place a Traits UI `Item` into a `Bound` layout.

    The automatically-added traits are only those for `QWidget`, not whatever
    widget the root control of the `Item` may turn out to be. This
    :class:`~.Binder` can only be used in the context of a :class:`~.Bound`
    layout because it needs to be specially recognized and initialized.
    """
    qclass = QtGui.QWidget

    #: The Traits UI Item to display. Any label is ignored.
    item = Instance(Item)

    # The `UI` object of the whole `View`.
    _ui = Any()

    # The initialized editor.
    _editor = Any()

    def __init__(self, item=None, **traits):
        super(TraitsUI, self).__init__(item=item, **traits)

    def initialize_item(self, ui):
        """ Initialize the item using the Traits UI `UI` object.
        """
        item = self.item
        name = item.name
        object = eval(item.object_, globals(), ui.context)
        trait = object.base_trait(name)

        # Get the editor factory associated with the Item:
        editor_factory = item.editor
        if editor_factory is None:
            editor_factory = trait.get_editor().trait_set(**item.editor_args)

            # If still no editor factory found, use a default text editor:
            if editor_factory is None:
                from traitsui.qt4.text_editor import ToolkitEditorFactory
                editor_factory = ToolkitEditorFactory()

            # If the item has formatting traits set them in the editor
            # factory:
            if item.format_func is not None:
                editor_factory.format_func = item.format_func

            if item.format_str != '':
                editor_factory.format_str = item.format_str

            # If the item has an invalid state extended trait name, set it
            # in the editor factory:
            if item.invalid != '':
                editor_factory.invalid = item.invalid

        # Create the requested type of editor from the editor factory:
        factory_method = getattr(editor_factory, item.style + '_editor')
        editor = factory_method(ui, object, name, item.tooltip, None)
        editor.trait_set(
            item=item,
            object_name=item.object,
        )
        self._editor = editor
        self._ui = ui

    def construct(self):
        item = self.item
        editor = self._editor
        ui = self._ui
        # Tell the editor to actually build the editing widget.  Note that
        # "inner" is a layout.  This shouldn't matter as individual editors
        # shouldn't be using it as a parent anyway.  The important thing is
        # that it is not None (otherwise the main TraitsUI code can change
        # the "kind" of the created UI object).
        inner = QtGui.QVBoxLayout()
        editor.prepare(inner)
        control = editor.control

        if item.style_sheet:
            control.setStyleSheet(item.style_sheet)

        # Set the initial 'enabled' state of the editor from the factory:
        editor.enabled = editor.factory.enabled

        # Give the editor focus if it requested it:
        if item.has_focus:
            control.setFocus()

        # We ignore the geometry options of the Item. At least for now.

        # Bind the editor into the UIInfo object name space so it can be
        # referred to by a Handler while the user interface is active:
        id = item.id or item.name
        ui.info.bind(id, editor, item.id)

        ui._scrollable |= editor.scrollable

        # Also, add the editors to the list of editors used to construct
        # the user interface:
        ui._editors.append(editor)

        # If the editor is conditionally visible, add the visibility
        # 'expression' and the editor to the UI object's list of monitored
        # objects:
        if item.visible_when != '':
            ui.add_visible(item.visible_when, editor)

        # If the editor is conditionally enabled, add the enabling
        # 'expression' and the editor to the UI object's list of monitored
        # objects:
        if item.enabled_when != '':
            ui.add_enabled(item.enabled_when, editor)

        self.qobj = control


class QtBoundEditor(Editor):
    """ Qt implementation of the ``BoundEditor``.
    """

    #: The ``Binder`` object being displayed.
    binder = Instance(Binder)

    def __init__(self, parent, **traits):
        """ Initializes the editor object.
        """
        HasPrivateTraits.__init__(self, **traits)
        self.old_value = Undefined

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
        widget.
        """
        binder = self.factory.binder
        for child in binder:
            if isinstance(child, TraitsUI):
                child.initialize_item(self.ui)
        binder.construct()
        binder.configure()
        self.binder = binder
        context = self._get_context()
        for button_group in self.factory.button_groups.values():
            button_group.construct()
            button_group.configure()
            button_group.add_buttons_from_context(context)
        for binding in self.factory.bindings:
            binding.bind(binder, context)
        if self.factory.configure is not None:
            self.factory.configure(binder, context)
        if isinstance(binder.qobj, QtGui.QLayout):
            self.control = QtGui.QWidget()
            self.control.setLayout(binder.qobj)
        else:
            self.control = binder.qobj
        if self.factory.stylesheet is not None:
            self.control.setStyleSheet(self.factory.stylesheet)

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        if self.ui is None:
            return

        for binding in self.factory.bindings:
            binding.unbind()

        self.binder.dispose()
        self.binder = None

        # Break linkages to references we no longer need:
        self.object = self.ui = self.item = self.factory = self.control = \
            self.label_control = self.old_value = self._context_object = None

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
        editor.

        No-op in our case.
        """
        pass

    def _get_context(self):
        """ Return a context for evaluating binding expressions.
        """
        context = self.ui.context.copy()
        context.update(self.factory.button_groups)
        context.update(self.factory.extra_context)
        # Add any binders with IDs to the context.
        for child in self.binder:
            if child.id:
                context[child.id] = child
        return context


class BoundEditor(EditorFactory):
    """ Editor for Binder-wrapped Qt widgets.
    """

    #: The Binder instance.
    binder = Instance(Binder)

    #: The list of ``Bindings``.
    bindings = List(Instance(Binding))

    #: Any extra data to put into the context.
    extra_context = Dict()

    #: A function to call to configure the Binder. It takes two arguments, the
    #: root Binder and the dictionary that is used as the context for
    #: evaluating the bindings.
    configure = Callable()

    #: The Qt stylesheet to apply to the root control.
    stylesheet = Either(Str, None)

    #: Mapping of names to `ButtonGroups`.
    button_groups = Dict(Str, Instance(ButtonGroup))

    def _get_simple_editor_class(self):
        return QtBoundEditor

    def _get_custom_editor_class(self):
        return QtBoundEditor

    def _get_text_editor_class(self):
        return QtBoundEditor

    def _get_readonly_editor_class(self):
        return QtBoundEditor


class Bound(Item):
    """ Convenience ``Item`` for placing a ``Binder`` in a ``View``.
    """

    def __init__(self, binder, *bindings, **kwds):
        extra_context = kwds.pop('extra_context', {})
        if 'label' not in kwds:
            kwds.setdefault('show_label', False)
        configure = kwds.pop('configure', None)
        stylesheet = kwds.pop('stylesheet', None)
        button_groups = kwds.pop('button_groups', {})
        # FIXME: find a better workaround for using `trait_modified`.
        # We use it because Traits UI expects *a* trait here. This is one that
        # is unlikely to appear elsewhere in Traits UIs. Fortunately, it is one
        # that every `HasTraits` class already has, and it is an `Event`, which
        # prevents Traits UI from trying to get its value.
        parsed_bindings = list(map(Binding.parse, bindings))
        super(Bound, self).__init__(
            value='trait_modified',
            editor=BoundEditor(
                binder=binder,
                bindings=parsed_bindings,
                extra_context=extra_context,
                configure=configure,
                stylesheet=stylesheet,
                button_groups=button_groups,
            ),
            **kwds)

    def __repr__(self):
        lines = [
            '{0.__name__}('.format(type(self)),
            '    {0!r},'.format(self.editor.binder),
        ]
        for binding in self.editor.bindings:
            if type(binding).__str__ is not object.__str__:
                lines.append("    '{0}',".format(binding))
            else:
                lines.append('    {0!r},'.format(binding))
        if self.editor.configure:
            lines.append('    configure={0.configure!r},'.format(self.editor))
        if self.editor.extra_context:
            lines.append('    extra_context={0.extra_context!r},'.format(
                self.editor))
        if self.editor.button_groups:
            lines.append('    button_groups={0.button_groups!r},'.format(
                self.editor))
        if self.show_label:
            lines.append('    label={0.label!r},'.format(self))
        lines.append(')')
        return '\n'.join(lines)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2014-2016, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------


import unittest

from ..qt import QtCore, QtGui
from ..raw_widgets import BasicGridLayout, GroupBox, HBoxLayout, Label, \
    LineEdit, Object, VBoxLayout, Widget, binder_registry
from ..testing import BaseTestWithGui


class TestBoxLayout(BaseTestWithGui, unittest.TestCase):

    def test_configure_nested_layout(self):
        # Regression test for a bug that prevented layouts from being nested.
        nested_layout = VBoxLayout(HBoxLayout(), VBoxLayout())
        nested_layout.construct()
        nested_layout.configure()
        content = [
            nested_layout.qobj.itemAt(idx)
            for idx in range(nested_layout.qobj.count())
        ]
        self.assertEqual(len(content), 2)
        self.assertIsInstance(content[0], QtGui.QHBoxLayout)
        self.assertIsInstance(content[1], QtGui.QVBoxLayout)

    def test_widget_inside_nested_layout_has_parent(self):
        # Regression test for a bug that prevented a widget inside nested
        # layouts from begin displayed.
        label = Label()
        box = GroupBox(VBoxLayout(VBoxLayout(label)))
        box.construct()
        box.configure()
        self.assertIs(label.qobj.parent(), box.qobj)


class TestBinderRegistry(unittest.TestCase):

    def test_lookup_object(self):
        self.assertIs(binder_registry.lookup_by_type(QtCore.QObject),
                      Object)

    def test_lookup_widget(self):
        self.assertIs(binder_registry.lookup_by_type(QtGui.QWidget),
                      Widget)

    def test_lookup_grid_layout(self):
        self.assertIs(binder_registry.lookup_by_type(QtGui.QGridLayout),
                      BasicGridLayout)


class TestGroupBox(BaseTestWithGui, unittest.TestCase):

    def test_group_box_can_be_childless(self):
        with self.constructed(GroupBox()) as box:
            self.assertIsInstance(box.qobj, QtGui.QGroupBox)

    def test_groupbox_alignment_works(self):
        # PySide has a bug. Ensure that we work around it.
        with self.constructed(GroupBox()) as box:
            # In particular, combined flags like this don't seem to work
            # through the meta_prop mechanism.
            new_alignment = QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop
            # Test getting.
            orig = box.alignment
            self.assertNotEqual(orig, new_alignment)
            # Test setting.
            box.alignment = new_alignment
            # Sometimes the flag comes back out as integers, which don't
            # compare equal to the flag object.
            self.assertEqual(int(box.alignment), int(new_alignment))
            # But the alignment() getter method should always return the
            # expected type.
            self.assertEqual(box.qobj.alignment(), new_alignment)


class TestQLineEdit(BaseTestWithGui, unittest.TestCase):

    def test_slots_with_arguments_work(self):
        with self.constructed(LineEdit()) as le:
            le.textEdited = u'text'



#------------------------------------------------------------------------------
#
#  Copyright (c) 2014-2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------

import unittest

from traits.testing.unittest_tools import UnittestTools

from ..qt import QtGui
from ..testing import BaseTestWithGui
from ..widgets import FloatSlider, RangeSlider, TextField, EditableComboBox


class TestTextField(unittest.TestCase, BaseTestWithGui, UnittestTools):

    def test_traits(self):
        field = TextField()
        self.assertEqual(field.value, u'')
        with self.assertTraitChanges(field, 'value', count=1):
            # The value trait should always fire a notification, even if the
            # value is not different.
            field.value = u''

    def test_validity_in_auto_mode(self):
        with self.constructed(TextField(mode='auto')) as field:
            field.validator = QtGui.QIntValidator(30, 50)

            # Set the text field in an invalid state.
            field.textEdited = '10'
            self.assertEqual(field.value, '10')
            self.assertEqual(field.valid, False)

            # In 'auto' mode, the value and the validity should update
            # automatically even before the user presses "Enter".
            field.textEdited = '40'
            self.assertEqual(field.value, '40')
            self.assertEqual(field.valid, True)


class TestRangeSlider(unittest.TestCase, BaseTestWithGui):

    def test_initialization(self):
        # With a random hash seed, this would fail randomly if we didn't take
        # special care in RangeSlider.__init__()
        inner = FloatSlider()
        outer = RangeSlider(slider=inner, range=(-10.0, 10.0))
        self.assertEqual(inner.range, (-10.0, 10.0))
        self.assertIs(outer.slider, inner)

    def test_field_format_func(self):
        slider = RangeSlider(field_format_func=u'+{0}'.format)
        slider.slider.value = 10
        self.assertEqual(slider.field.text, u'+10')
        slider.value = 20
        self.assertEqual(slider.field.text, u'20')

    def test_field_text_edited(self):
        with self.constructed(RangeSlider()) as slider:
            self.assertEqual(slider.value, 0)
            self.assertEqual(slider.slider.value, 0)
            slider.field.text = '20'
            slider.field.editingFinished = True
            self.assertEqual(slider.value, 20)
            self.assertEqual(slider.slider.value, 20)

    def test_validator_states(self):
        # When the validator is Invalid or Intermediate, the value of the
        # RangeSlider should not change.
        range_slider = RangeSlider(range=(30, 60), value=50)
        with self.constructed(range_slider):
            slider = range_slider.slider
            field = range_slider.field

            self.assertEqual(slider.value, 50)
            self.assertEqual(range_slider.value, 50)

            # State is Intermediate.
            field.text = '1'
            field.editingFinished = True
            self.assertEqual(slider.value, 50)
            self.assertEqual(range_slider.value, 50)

            # State is Invalid.
            field.text = '-1'
            field.editingFinished = True
            self.assertEqual(slider.value, 50)
            self.assertEqual(range_slider.value, 50)


class TestEditableComboBox(unittest.TestCase, BaseTestWithGui):

    def test_overloaded_signals(self):
        with self.constructed(EditableComboBox()) as combo:
            combo.values = [(0, 'zero'), (1, 'one'), (2, 'two')]
            combo.setCurrentIndex(1)
            self.assertEqual(combo.value, 1)

            received = []

            def slot(value):
                received.append(value)

            combo.on_trait_change(slot, 'currentIndexChanged_int')
            combo.setCurrentIndex(2)
            self.assertEqual(received, [2])
            received[:] = []
            combo.on_trait_change(slot, 'currentIndexChanged_int', remove=True)

            combo.on_trait_change(slot, 'currentIndexChanged_QString')
            combo.setCurrentIndex(0)
            self.assertEqual(received, ['zero'])
            received[:] = []
            combo.on_trait_change(slot, 'currentIndexChanged_QString',
                                  remove=True)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2014-2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------

import unittest

import six

from traits.api import Bool, Instance, pop_exception_handler, \
    push_exception_handler

from ..binder import Binder, Composite, Default, QtDynamicProperty, \
    QtGetterSetter, QtProperty, QtSignal, QtSlot, Rename
from ..qt import QtCore, QtGui, qt_api


class TestBinder(unittest.TestCase):

    def setUp(self):
        # Start up a QApplication if needed.
        self.app = QtGui.QApplication.instance()
        if self.app is None:
            self.app = QtGui.QApplication([])

        push_exception_handler(reraise_exceptions=True)

        # We make a new class here to test that the traits get added only at
        # the right time.
        # WARNING: If anyone instantiates Binder() itself, this subclass will
        # see its added traits, and the test will fail.
        class Object(Binder):
            qclass = QtCore.QObject

            x = QtDynamicProperty(10)

        self.Object = Object

    def tearDown(self):
        pop_exception_handler()

    def test_object_traits(self):
        # No QtTraits made before instantiation.
        previous_traits = {k: v.trait_type
                           for k, v in self.Object.class_traits().items()}
        self.assertNotIn('destroyed', previous_traits)

        self.Object()
        traits = {k: v.trait_type
                  for k, v in self.Object.class_traits().items()}
        self.assertIsInstance(traits['deleteLater'], QtSlot)
        self.assertIsInstance(traits['destroyed'], QtSignal)
        self.assertIsInstance(traits['destroyed_QObject'], QtSignal)
        self.assertIsInstance(traits['objectName'], QtProperty)
        self.assertIsInstance(traits['parent'], QtGetterSetter)
        self.assertIsInstance(traits['x'], QtDynamicProperty)

    def test_dynamic_properties(self):
        obj = self.Object()
        # QtDynamicProperties work correctly before qobj is provided.
        self.assertEqual(obj.x, 10)
        obj.x = 20
        self.assertEqual(obj.x, 20)

        obj.construct()
        qobj = obj.qobj
        obj.configure()

        # QtDynamicProperties work after qobj is provided.
        self.assertEqual(obj.x, 20)
        self.assertEqual(qobj.property('x'), 20)
        obj.x = 30
        self.assertEqual(obj.x, 30)
        self.assertEqual(qobj.property('x'), 30)

        obj.dispose()

    def test_static_properties(self):
        obj = self.Object()
        with self.assertRaises(AttributeError):
            obj.objectName
        obj.objectName = 'object'
        self.assertEqual(obj.objectName, 'object')

        obj.construct()
        qobj = obj.qobj
        obj.configure()

        self.assertEqual(obj.objectName, 'object')
        self.assertEqual(qobj.objectName(), 'object')
        obj.objectName = 'foo'
        self.assertEqual(obj.objectName, 'foo')
        self.assertEqual(qobj.objectName(), 'foo')

        obj.dispose()

    def test_getter_setter(self):
        obj = self.Object()
        with self.assertRaises(AttributeError):
            obj.parent
        parent_qobj = QtCore.QObject()
        other_parent = QtCore.QObject()
        obj.parent = parent_qobj
        self.assertIs(obj.parent, parent_qobj)

        obj.construct()
        qobj = obj.qobj
        obj.configure()

        self.assertIs(obj.parent, parent_qobj)
        self.assertIs(qobj.parent(), parent_qobj)
        obj.parent = other_parent
        self.assertIs(obj.parent, other_parent)
        self.assertIs(qobj.parent(), other_parent)

        obj.dispose()

    def test_slot(self):
        obj = self.Object()
        with self.assertRaises(AttributeError):
            obj.deleteLater

        obj.deleteLater = True
        obj.construct()
        qobj = obj.qobj
        obj.configure()

        self.assertEqual(obj.deleteLater, qobj.deleteLater)
        # FIXME: Test setting afterwards.
        # FIXME: Test other signatures.

        obj.dispose()

    def test_signal(self):
        obj = self.Object()
        with self.assertRaises(AttributeError):
            obj.destroyed

        obj.destroyed = True
        qobj = QtCore.QObject()
        received = []

        def handler():
            received.append(True)

        qobj.destroyed.connect(handler)

        obj.qobj = qobj
        obj.configure()

        self.assertEqual(received, [True])
        obj.destroyed = True
        self.assertEqual(received, [True, True])

        obj.dispose()

    def test_delayed_connection(self):
        obj = self.Object()
        received = []

        def handler(new):
            received.append(new)

        obj.on_trait_change(handler, 'destroyed')
        with self.assertRaises(AttributeError):
            obj.destroyed

        obj.construct()
        qobj = obj.qobj
        obj.configure()

        self.assertEqual(type(obj.destroyed), type(qobj.destroyed))

        # Check delayed signal.
        qobj.destroyed[QtCore.QObject].emit(qobj)

        if qt_api.startswith('pyqt'):
            # When there are 2 overloads for a signal and one of the is
            # argumentless, we actually get the single-argument version regardless.
            # This is actually typically what we want, but I'm noting it because
            # this behavior did change when we redid the signal implementation to
            # conform to the limitations of the PyQt5 signal implementation.
            six.assertCountEqual(self, received, [qobj])
        elif qt_api.startswith('pyside'):
            # But PySides don't, and I haven't found a good way to rationalize
            # the two.
            six.assertCountEqual(self, received, [()])

        # No signal after removal.
        received[:] = []
        obj.on_trait_change(handler, 'destroyed', remove=True)

        qobj.destroyed[QtCore.QObject].emit(qobj)
        self.assertEqual(received, [])

        # No signal after disposal.
        obj.on_trait_change(handler, 'destroyed')
        obj.dispose()
        qobj.destroyed[QtCore.QObject].emit(qobj)
        self.assertEqual(received, [])

    def test_composite(self):
        class LineEdit(Binder):
            qclass = QtGui.QLineEdit

            _configured = Bool(False)
            _disposed = Bool(False)

            def configure(self):
                self._configured = True

            def dispose(self):
                super(LineEdit, self).dispose()
                self._disposed = True

        class ComboBox(Composite):
            qclass = QtGui.QComboBox
            lineEdit = Instance(LineEdit)

            def configure(self):
                self.editable = True
                self.lineEdit = LineEdit(qobj=self.qobj.lineEdit())
                super(ComboBox, self).configure()

        cb = ComboBox()
        cb.construct()
        cb.configure()
        self.assertEqual(list(cb), [cb, cb.lineEdit])
        self.assertEqual(cb.child_binders, [cb.lineEdit])
        self.assertTrue(cb.lineEdit._configured)
        self.assertFalse(cb.lineEdit._disposed)
        cb.dispose()
        self.assertTrue(cb.lineEdit._disposed)

    def test_default_rename(self):
        class Widget(Binder):
            qclass = QtGui.QWidget

            accessibleName = Default(u'blah')
            what_is_this = Rename('whatsThis', default=u'Foo')
            object_name = Rename('objectName')
            the_layout = Rename('layout')
            py_raise = Rename('raise_')

        class SubWidget(Widget):
            pass

        # Test the subclass first so it can't just inherit the traits of the
        # parent class.
        for cls in [SubWidget, Widget]:
            w = cls()
            w.construct()
            self.assertEqual(w.accessibleName, u'blah')
            self.assertEqual(w.qobj.accessibleName(), u'blah')
            self.assertEqual(w.what_is_this, u'Foo')
            self.assertEqual(w.qobj.whatsThis(), u'Foo')
            self.assertEqual(w.object_name, w.qobj.objectName())
            self.assertEqual(w.the_layout, None)
            self.assertEqual(w.py_raise, w.qobj.raise_)
            self.assertNotIn('raise', sorted(w.class_traits()))
            with self.assertRaises(AttributeError):
                w.whatsThis
            with self.assertRaises(AttributeError):
                w.objectName
            with self.assertRaises(AttributeError):
                w.layout
            w.dispose()

            w = cls(accessibleName=u'not-blah', what_is_this=u'Bar')
            w.construct()
            self.assertEqual(w.accessibleName, u'not-blah')
            self.assertEqual(w.qobj.accessibleName(), u'not-blah')
            self.assertEqual(w.what_is_this, u'Bar')
            self.assertEqual(w.qobj.whatsThis(), u'Bar')
            w.dispose()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2014-2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------

import unittest

import six

from traits.api import Any, Event, HasTraits, NO_COMPARE, \
    pop_exception_handler, push_exception_handler, Instance

from ..binder import Binder
from ..binding import Binding, Factory, PulledFrom, PushedTo, SetOnceTo, \
    SyncedWith, find_ext_attrs


class DummyBinder(Binder):
    x = Any(comparison_mode=NO_COMPARE)
    child = Instance(HasTraits)


class DummyModel(HasTraits):
    y = Any(10, comparison_mode=NO_COMPARE)
    event = Event()


class TestBinding(unittest.TestCase):

    def setUp(self):
        self.binder = DummyBinder()
        self.model = DummyModel()
        self.blah = DummyBinder()
        self.context = dict(
            object=self.model,
            blah=self.blah,
        )
        push_exception_handler(lambda *args, **kwds: None,
                               reraise_exceptions=True)

    def tearDown(self):
        pop_exception_handler()

    def test_find_ext_attr(self):
        pairs = [
            ('object', []),  # no dots
            ('object.foo', ['object.foo']),
            ('object.foo.bar', ['object.foo.bar']),
            ('object.foo + 10', ['object.foo']),
            ('object.foo + str(int)', ['object.foo']),
            ('object.foo + str(int)', ['object.foo']),
            ('object.foo + handler.bar', ['object.foo', 'handler.bar']),
            ('object.foo + handler.func(10)', ['object.foo', 'handler.func']),
            ('object.foo + "ohm.m"', ['object.foo']),
            ('object.foo + "ohm.m".format.__name__', ['object.foo']),
            ('(object.foo).text()', ['object.foo']),
        ]
        for expr, ext_attrs in pairs:
            found = find_ext_attrs(expr)
            six.assertCountEqual(self, found, ext_attrs)

    def test_parse_binding(self):
        pairs = [
            ('text = object.text',
             SetOnceTo('text', 'object.text')),
            ('text=object.text',
             SetOnceTo('text', 'object.text')),
            ('checked = object.text == "blah"',
             SetOnceTo('checked', 'object.text == "blah"')),
            ('text = object.func(value="blah")',
             SetOnceTo('text', 'object.func(value="blah")')),
            ('text << object.text',
             PulledFrom('text', 'object.text')),
            ('text<<object.text',
             PulledFrom('text', 'object.text')),
            ('value << object.value << 2',
             PulledFrom('value', 'object.value << 2')),
            ('value << object.value >> 2',
             PulledFrom('value', 'object.value >> 2')),
            ('text >> object.text',
             PushedTo('text', 'object.text')),
            ('text>>object.text',
             PushedTo('text', 'object.text')),
            ('value >> object.value >> 2',
             PushedTo('value', 'object.value >> 2')),
            ('value >> object.value << 2',
             PushedTo('value', 'object.value << 2')),
            ('value := object.value',
             SyncedWith('value', 'object.value')),
            ('value:=object.value',
             SyncedWith('value', 'object.value')),
            ('lineEdit.text << object.text',
             PulledFrom('lineEdit.text', 'object.text')),
        ]
        for text, binding in pairs:
            parsed = Binding.parse(text)
            self.assertEqual(parsed, binding)
            self.assertEqual(hash(parsed), hash(binding))
            self.assertFalse(parsed != binding)
            self.assertFalse(parsed == 10)
            self.assertIs(Binding.parse(binding), binding)
        factory = Factory('text', lambda: 'foo')
        self.assertIs(Binding.parse(factory), factory)
        for bad in [None, ('value', 'object.value')]:
            with self.assertRaises(TypeError):
                Binding.parse(bad)

    def test_set_once_to(self):
        binding = SetOnceTo('x', 'object.y')
        binding.bind(self.binder, self.context)
        self.assertEqual(self.binder.x, 10)
        self.model.y = 20
        self.assertEqual(self.binder.x, 10)
        binding.unbind()
        self.model.y = 30
        self.assertEqual(self.binder.x, 10)

    def test_factory(self):
        binding = Factory('x', lambda: self.model.y)
        binding.bind(self.binder, self.context)
        self.assertEqual(self.binder.x, 10)
        self.model.y = 20
        self.assertEqual(self.binder.x, 10)
        binding.unbind()
        self.model.y = 30
        self.assertEqual(self.binder.x, 10)

    def test_pulled_from_trait(self):
        binding = PulledFrom('x', 'object.y')
        binding.bind(self.binder, self.context)
        self.assertEqual(self.binder.x, 10)
        self.model.y = 20
        self.assertEqual(self.binder.x, 20)
        binding.unbind()
        self.model.y = 30
        self.assertEqual(self.binder.x, 20)

    def test_pulled_from_expression(self):
        binding = PulledFrom('x', 'object.y + 5')
        binding.bind(self.binder, self.context)
        self.assertEqual(self.binder.x, 15)
        self.model.y = 20
        self.assertEqual(self.binder.x, 25)
        binding.unbind()
        self.model.y = 30
        self.assertEqual(self.binder.x, 25)

    def test_pulled_from_event(self):
        binding = PulledFrom('x', 'object.event')
        binding.bind(self.binder, self.context)
        self.assertIsNone(self.binder.x)
        self.model.event = 10
        self.assertEqual(self.binder.x, 10)
        binding.unbind()
        self.model.event = 20
        self.assertEqual(self.binder.x, 10)

    def test_pushed_to(self):
        binding = PushedTo('x', 'object.y')
        binding.bind(self.binder, self.context)
        # PushedTo does not initialize the model.
        self.assertEqual(self.model.y, 10)
        self.binder.x = 20
        self.assertEqual(self.model.y, 20)
        binding.unbind()
        self.binder.x = 30
        self.assertEqual(self.model.y, 20)

    def test_synced_with(self):
        binding = SyncedWith('x', 'object.y')
        binding.bind(self.binder, self.context)
        self.assertEqual(self.binder.x, 10)
        self.model.y = 20
        self.assertEqual(self.binder.x, 20)
        self.binder.x = 25
        self.assertEqual(self.model.y, 25)
        binding.unbind()
        self.model.y = 30
        self.assertEqual(self.binder.x, 25)
        self.binder.x = 35
        self.assertEqual(self.model.y, 30)

    def test_child(self):
        self.binder.child = DummyBinder()
        binding = SyncedWith('child.x', 'object.y')
        binding.bind(self.binder, self.context)
        self.assertIsNone(self.binder.x)
        self.assertEqual(self.binder.child.x, 10)
        self.model.y = 20
        self.assertEqual(self.binder.child.x, 20)
        self.binder.child.x = 25
        self.assertEqual(self.model.y, 25)
        binding.unbind()
        self.model.y = 30
        self.assertEqual(self.binder.child.x, 25)
        self.binder.child.x = 35
        self.assertEqual(self.model.y, 30)

    def test_binder_in_context(self):
        binding = PulledFrom('blah.x', 'object.y')
        binding.bind(self.binder, self.context)
        self.assertEqual(self.blah.x, 10)
        self.model.y = 20
        self.assertEqual(self.blah.x, 20)
        binding.unbind()
        self.model.y = 30
        self.assertEqual(self.blah.x, 20)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2014-2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------

import os
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))  # noqa

from traits.api import Enum, HasTraits, Unicode
from traitsui.api import Item, View

from qt_binder.api import Bound
from qt_binder.widgets import TextField, EnumDropDown


class Model(HasTraits):
    mode = Enum('auto', 'enter')
    value = Unicode()

    def default_traits_view(self):
        traits_view = View(
            Bound(
                EnumDropDown(
                    values=[('auto', u'Automatic'),
                            ('enter', u'On Enter')]),
                'value := object.mode',
                label=u'Mode:',
            ),
            Item('value', style='readonly'),
            Bound(
                TextField(),
                'mode := object.mode',
                'value := object.value',
                label=u'Text:',
            ),
        )
        return traits_view


def main():
    m = Model(value=u'Something')
    m.configure_traits()

if __name__ == '__main__':
    main()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2014-2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------

from __future__ import print_function

import os
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))  # noqa

from scimath.units.api import unit_parser
from traits.api import Bool, HasStrictTraits, List, Property, Str, Tuple, \
    Unicode, cached_property
from traitsui.api import View, VGroup, Controller

from qt_binder.api import Bound
from qt_binder.widgets import EditableComboBox, TextField
from qt_binder.raw_widgets import CheckBox, HBoxLayout


def unit_string_is_valid(units):
    return unit_parser.parse_unit(units).valid


def units_compatible(u0, u1):
    return (unit_parser.parse_unit(u0).derivation ==
            unit_parser.parse_unit(u1).derivation)


class Model(HasStrictTraits):
    input_units = Str()
    input_units_valid = Property(Bool, depends_on=['input_units'])
    convert_units = Bool(False)
    output_units = Str()
    output_units_valid = Property(Bool, depends_on=['input_units',
                                                    'convert_units',
                                                    'output_units'])

    @cached_property
    def _get_input_units_valid(self):
        return unit_string_is_valid(self.input_units)

    @cached_property
    def _get_output_units_valid(self):
        valid = True
        if self.convert_units:
            valid = unit_string_is_valid(self.output_units)
            if valid:
                valid = units_compatible(self.input_units, self.output_units)
        return valid


class ModelUI(Controller):

    output_units_recommendations = Property(List(Tuple(Str, Unicode)),
                                            depends_on=['model.input_units'])

    all_units_recommendations = List([
        ('m', u'Depth'),
        ('cm', u'Width'),
        ('ms', u'Time'),
        ('none', u'None'),
        ('radians', u'Radians'),
        ('deg', u'Degrees'),
    ])

    def default_traits_view(self):
        traits_view = View(
            VGroup(
                Bound(
                    TextField(id='input_units'),
                    'valid << object.input_units_valid',
                    'value := object.input_units',
                    label=u'Input units:',
                ),
                Bound(
                    HBoxLayout(
                        CheckBox(id='check'),
                        EditableComboBox(id='combo'),
                    ),
                    'combo.lineEdit.valid << object.output_units_valid',
                    'combo.values << handler.output_units_recommendations',
                    'combo.enabled << object.convert_units',
                    'check.checked := object.convert_units',
                    'combo.value >> object.output_units',
                    label=u'Convert units:',
                ),
            ),
            buttons=['OK'],
        )
        return traits_view

    def _get_output_units_recommendations(self):
        recommendations = []
        if unit_string_is_valid(self.model.input_units):
            for unit, description in self.all_units_recommendations:
                if units_compatible(self.model.input_units, unit):
                    recommendations.append(
                        (unit,
                         u'{0} \N{EM DASH} {1}'.format(description, unit)))
        return recommendations


def main():
    m = Model(input_units='m')
    u = ModelUI(m)
    u.configure_traits()
    print('Input: {0.input_units}'.format(m))
    print('Convert: {0.convert_units}'.format(m))
    print('Output: {0.output_units}'.format(m))
    print('Valid: {0.output_units_valid}'.format(m))

if __name__ == '__main__':
    main()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2014-2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------
""" RangeEditor replacement.
"""

import os
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))  # noqa

from traits.api import HasTraits, Float, Int
from traitsui.api import View

from qt_binder.api import Bound
from qt_binder.qt import QtGui
from qt_binder.raw_widgets import FormLayout
from qt_binder.widgets import FloatSlider, LogSlider, RangeSlider


class Model(HasTraits):
    int_value = Int(0)
    float_value = Float(0.0)
    log_value = Float(1.0)

    def default_traits_view(self):
        traits_view = View(
            Bound(
                FormLayout(
                    (u'Integer:', RangeSlider(
                        id='int_slider',
                        range=(-100, 100))),
                    (u'Float:', RangeSlider(
                        id='float_slider',
                        slider=FloatSlider(),
                        range=(-10.0, 10.0),
                        field_format_func=u'{0:.2f}'.format)),
                    (u'Log:', RangeSlider(
                        id='log_slider',
                        slider=LogSlider(),
                        range=(0.5, 2.0),
                        field_format_func=u'{0:.6f}'.format)),
                    fieldGrowthPolicy=QtGui.QFormLayout.ExpandingFieldsGrow,
                ),
                'int_slider.value := object.int_value',
                'float_slider.value := object.float_value',
                'log_slider.value := object.log_value',
                # Make the labels align nicely by fixing their widths using
                # stylesheets.
                stylesheet=(u'*[binder_class="RangeSlider"] > .QLabel '
                            u'{min-width: 40px; max-width: 40px;}'),
            ),
            resizable=True,
            title=u'Range Sliders',
        )
        return traits_view


def main():
    m = Model()
    m.configure_traits()

if __name__ == '__main__':
    main()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2014-2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------

import six

from traits.api import CList, Either, Instance, Int, List, Property, Str, \
    TraitError, Tuple, Unicode, on_trait_change

from .binder import Binder, Composite, NChildren
from .qt import QtCore, QtGui
from .qt.QtCore import Qt
from .type_registry import TypeRegistry


#: The global registry mapping PySide/PyQt types to their Binders.
binder_registry = TypeRegistry()


class Object(Binder):
    qclass = QtCore.QObject


class Widget(Binder):
    qclass = QtGui.QWidget


class DialogButtonBox(Binder):
    qclass = QtGui.QDialogButtonBox


class DesktopWidget(Binder):
    qclass = QtGui.QDesktopWidget


class LineEdit(Binder):
    qclass = QtGui.QLineEdit


class ComboBox(Composite):
    """ Customized to exposed the line-edit widget as a child
    :class:`~.Binder`.
    """
    qclass = QtGui.QComboBox

    lineEdit_class = LineEdit
    lineEdit = Instance(LineEdit)

    @on_trait_change('editable,qobj')
    def _on_editable(self):
        qobj = self.qobj
        if qobj is not None and self.editable:
            self.lineEdit = self.lineEdit_class(qobj=qobj.lineEdit())
            self.lineEdit.configure()
        else:
            if self.lineEdit is not None:
                self.lineEdit.dispose()
            self.lineEdit = None


class FontComboBox(ComboBox):
    qclass = QtGui.QFontComboBox


class CalendarWidget(Binder):
    qclass = QtGui.QCalendarWidget


class PushButton(Binder):
    qclass = QtGui.QPushButton


class CommandLinkButton(Binder):
    qclass = QtGui.QCommandLinkButton


class CheckBox(Binder):
    qclass = QtGui.QCheckBox


class ToolButton(Binder):
    qclass = QtGui.QToolButton


class RadioButton(Binder):
    qclass = QtGui.QRadioButton


class WizardPage(Binder):
    qclass = QtGui.QWizardPage


class Frame(Binder):
    qclass = QtGui.QFrame


class ToolBox(Binder):
    qclass = QtGui.QToolBox


class TreeView(Binder):
    qclass = QtGui.QTreeView


class TreeWidget(Binder):
    qclass = QtGui.QTreeWidget


class HeaderView(Binder):
    qclass = QtGui.QHeaderView


class TableView(Binder):
    qclass = QtGui.QTableView


class TableWidget(Binder):
    qclass = QtGui.QTableWidget


class ColumnView(Binder):
    qclass = QtGui.QColumnView


class ListView(Binder):
    qclass = QtGui.QListView


class UndoView(Binder):
    qclass = QtGui.QUndoView


class ListWidget(Binder):
    qclass = QtGui.QListWidget


class ScrollArea(Binder):
    qclass = QtGui.QScrollArea


class PlainTextEdit(Binder):
    qclass = QtGui.QPlainTextEdit


class TextEdit(Binder):
    qclass = QtGui.QTextEdit


class TextBrowser(Binder):
    qclass = QtGui.QTextBrowser


class MdiArea(Binder):
    qclass = QtGui.QMdiArea


class GraphicsView(Binder):
    qclass = QtGui.QGraphicsView


class StackedWidget(Binder):
    qclass = QtGui.QStackedWidget


class LCDNumber(Binder):
    qclass = QtGui.QLCDNumber


class Label(Binder):
    qclass = QtGui.QLabel


class ToolBar(Binder):
    qclass = QtGui.QToolBar


class RubberBand(Binder):
    qclass = QtGui.QRubberBand


class TabWidget(Binder):
    qclass = QtGui.QTabWidget


class StatusBar(Binder):
    qclass = QtGui.QStatusBar


class TabBar(Binder):
    qclass = QtGui.QTabBar


class SplitterHandle(Binder):
    qclass = QtGui.QSplitterHandle


class Slider(Binder):
    qclass = QtGui.QSlider


class Dial(Binder):
    qclass = QtGui.QDial


class ScrollBar(Binder):
    qclass = QtGui.QScrollBar


class DateTimeEdit(Binder):
    qclass = QtGui.QDateTimeEdit


class DateEdit(Binder):
    qclass = QtGui.QDateEdit


class TimeEdit(Binder):
    qclass = QtGui.QTimeEdit


class DoubleSpinBox(Binder):
    qclass = QtGui.QDoubleSpinBox


class SpinBox(Binder):
    qclass = QtGui.QSpinBox


class SplashScreen(Binder):
    qclass = QtGui.QSplashScreen


class SizeGrip(Binder):
    qclass = QtGui.QSizeGrip


class Dialog(Binder):
    qclass = QtGui.QDialog


class ColorDialog(Binder):
    qclass = QtGui.QColorDialog


class PrintDialog(Binder):
    qclass = QtGui.QPrintDialog


class PageSetupDialog(Binder):
    qclass = QtGui.QPageSetupDialog


class FileDialog(Binder):
    qclass = QtGui.QFileDialog


class Wizard(Binder):
    qclass = QtGui.QWizard


class ProgressDialog(Binder):
    qclass = QtGui.QProgressDialog


class PrintPreviewDialog(Binder):
    qclass = QtGui.QPrintPreviewDialog


class FontDialog(Binder):
    qclass = QtGui.QFontDialog


class MessageBox(Binder):
    qclass = QtGui.QMessageBox


class InputDialog(Binder):
    qclass = QtGui.QInputDialog


class ErrorMessage(Binder):
    qclass = QtGui.QErrorMessage


class ProgressBar(Binder):
    qclass = QtGui.QProgressBar


class PrintPreviewWidget(Binder):
    qclass = QtGui.QPrintPreviewWidget


class MenuBar(Binder):
    qclass = QtGui.QMenuBar


class Menu(Binder):
    qclass = QtGui.QMenu


class MdiSubWindow(Binder):
    qclass = QtGui.QMdiSubWindow


class MainWindow(Binder):
    qclass = QtGui.QMainWindow


class SingleChild(Composite):
    """ Base class for widgets that typically have just a single child.
    """
    qclass = QtGui.QWidget

    child = Instance(Binder)

    def __init__(self, child=None, **traits):
        super(SingleChild, self).__init__(child=child, **traits)

    def construct(self):
        if self.child is not None:
            self.child.construct()
        super(SingleChild, self).construct()

    def configure(self):
        if self.child is not None:
            if isinstance(self.child.qobj, QtGui.QWidget):
                self.child.qobj.setParent(self.qobj)
            elif isinstance(self.child.qobj, QtGui.QLayout):
                self.qobj.setLayout(self.child.qobj)
        super(SingleChild, self).configure()


class GroupBox(SingleChild):
    qclass = QtGui.QGroupBox


class FocusFrame(Binder):
    qclass = QtGui.QFocusFrame


class DockWidget(Binder):
    qclass = QtGui.QDockWidget


class Layout(NChildren):
    """ Base class for all `QLayouts`.
    """
    qclass = QtGui.QLayout

    def __init__(self, *children, **kwds):
        self.child_binders = list(children)
        super(Layout, self).__init__(**kwds)

    def construct(self):
        """ Build the QLayout.
        """
        for child in self.child_binders:
            child.construct()
        super(Layout, self).construct()


class Splitter(NChildren):
    """ A splitter widget for arbitrary numbers of children.
    """
    qclass = QtGui.QSplitter

    def __init__(self, *children, **kwds):
        self.child_binders = list(children)
        super(Splitter, self).__init__(**kwds)

    def construct(self):
        """ Build the QLayout.
        """
        for child in self.child_binders:
            child.construct()
        super(Layout, self).construct()

    def configure(self):
        super(StackedLayout, self).configure()
        qobj = self.qobj
        for child in self.child_binders:
            widget = child.qobj
            if isinstance(widget, QtGui.QLayout):
                widget = QtGui.QWidget()
                widget.setLayout(child.qobj)
            qobj.addWidget(widget)


class BoxLayout(Layout):
    """ Base class for box layouts.
    """
    qclass = QtGui.QBoxLayout

    def configure(self):
        super(BoxLayout, self).configure()
        qobj = self.qobj
        for child in self.child_binders:
            if isinstance(child.qobj, QtGui.QWidget):
                qobj.addWidget(child.qobj)
            elif isinstance(child.qobj, QtGui.QLayout):
                qobj.addLayout(child.qobj)


class VBoxLayout(BoxLayout):
    """ A vertical layout.
    """
    qclass = QtGui.QVBoxLayout


class HBoxLayout(BoxLayout):
    """ A horizontal layout.
    """
    qclass = QtGui.QHBoxLayout


class StackedLayout(Layout):
    """ A stacked layout.
    """
    qclass = QtGui.QStackedLayout

    def configure(self):
        super(StackedLayout, self).configure()
        qobj = self.qobj
        for child in self.child_binders:
            widget = child.qobj
            if isinstance(widget, QtGui.QLayout):
                widget = QtGui.QWidget()
                widget.setLayout(child.qobj)
            qobj.addWidget(widget)


class BasicGridLayout(Layout):
    """ An explicit grid layout without colspans or rowspans.

    The arguments are equal-length lists of `Binder` widgets, `unicode` labels,
    `(Binder, Qt.Alignment)` tuples, `(unicode, Qt.Alignment)` tuples, or
    `None` for an empty grid cell.
    """
    qclass = QtGui.QGridLayout

    #: List of lists of `Binders`, `unicode` labels, or `None`. Each list
    #: should have the same number of elements (i.e. the number of columns.
    rows = List(CList(Either(
        None,
        Instance(Binder),
        Unicode,
        Tuple(Instance(Binder), Either(Instance(Qt.AlignmentFlag),
                                       Instance(Qt.Alignment))),
        Tuple(Unicode, Either(Instance(Qt.AlignmentFlag),
                              Instance(Qt.Alignment))),
    )))

    #: The child ``Binder`` instances.
    child_binders = Property(List(Instance(Binder)))

    def __init__(self, *rows, **traits):
        rows = list(rows)
        all_ncols = set(len(row) for row in rows)
        if len(all_ncols) > 1:
            # FIXME: Validate on trait assignment, too.
            raise TraitError("Expected the same number of columns: "
                             "got {!r}".format(all_ncols))
        # Skip the Layout.__init__().
        super(Layout, self).__init__(rows=rows, **traits)

    def __repr__(self):
        args = ',\n  '.join(map(repr, self.rows))
        if self.id:
            args += ',\n  id={0.id!r}'.format(self)
        return '{0.__name__}(\n  {1})'.format(type(self), args)

    def configure(self):
        super(Layout, self).configure()
        qobj = self.qobj
        for irow, row in enumerate(self.rows):
            for icol, cell in enumerate(row):
                alignment = Qt.Alignment(0)
                if isinstance(cell, tuple):
                    cell, alignment = cell
                if isinstance(cell, six.string_types):
                    widget = QtGui.QLabel()
                    widget.setText(cell)
                    qobj.addWidget(widget, irow, icol, alignment)
                elif isinstance(cell, Binder):
                    if isinstance(cell.qobj, QtGui.QWidget):
                        qobj.addWidget(cell.qobj, irow, icol, alignment)
                    elif isinstance(cell.qobj, QtGui.QLayout):
                        qobj.addLayout(cell.qobj, irow, icol, alignment)
                    else:
                        raise TypeError("Expected a QWidget or QLayout: "
                                        "got {!r}".format(cell.qobj))
                # Ignore None.

    def _get_child_binders(self):
        children = []
        for row in self.rows:
            for cell in row:
                if isinstance(cell, Binder):
                    children.append(cell)
                elif isinstance(cell, tuple) and isinstance(cell[0], Binder):
                    children.append(cell[0])
        return children


class SpanGridLayout(Layout):
    """ Grid layout with spans.
    """
    qclass = QtGui.QGridLayout

    #: Items to add: `(Binder/unicode, row, col[, rowspan, colspan]
    #: [, alignment])`
    items = List(Either(
        Tuple(Either(Instance(Binder), Unicode), Int, Int),
        Tuple(Either(Instance(Binder), Unicode), Int, Int,
              Instance(Qt.AlignmentFlag)),
        Tuple(Either(Instance(Binder), Unicode), Int, Int, Int, Int),
        Tuple(Either(Instance(Binder), Unicode), Int, Int, Int, Int,
              Instance(Qt.AlignmentFlag)),
    ))

    #: The child ``Binder`` instances.
    child_binders = Property(List(Instance(Binder)))

    def __init__(self, *items, **traits):
        items = list(items)
        # Skip the Layout.__init__().
        super(Layout, self).__init__(items=items, **traits)

    def __repr__(self):
        args = ',\n  '.join(map(repr, self.items))
        if self.id:
            args += ',\n  id={0.id!r}'.format(self)
        return '{0.__name__}(\n  {1})'.format(type(self), args)

    def configure(self):
        super(Layout, self).configure()
        qobj = self.qobj
        for item in self.items:
            cell = item[0]
            if isinstance(cell, six.string_types):
                label = QtGui.QLabel()
                label.setText(cell)
                args = (label,) + item[1:]
            else:  # isiinstance(cell, Binder)
                args = (cell.qobj,) + item[1:]
            if isinstance(args[0], QtGui.QWidget):
                qobj.addWidget(*args)
            elif isinstance(args[0], QtGui.QLayout):
                qobj.addLayout(*args)
            else:
                raise TypeError("Expected a QWidget or QLayout: "
                                "got {!r}".format(args[0]))

    def _get_child_binders(self):
        children = []
        for item in self.items:
            if isinstance(item[0], Binder):
                children.append(item[0])
        return children


class FormLayout(Layout):
    """ Children are (label, widget) pairs.

    The label can be a ``unicode`` string or ``None``. The last item can be
    a single ``Binder`` to take up the whole space.
    """
    qclass = QtGui.QFormLayout

    #: The (label, widget) pairs.
    rows = List(Either(
        Tuple(Either(None, Unicode, Instance(Binder)), Instance(Binder)),
        Instance(Binder),
    ))

    #: The child ``Binder`` instances.
    child_binders = Property(List(Instance(Binder)))

    def __init__(self, *rows, **traits):
        rows = list(rows)
        # Skip the Layout.__init__().
        super(Layout, self).__init__(rows=rows, **traits)

    def configure(self):
        super(Layout, self).configure()
        qobj = self.qobj
        for row in self.rows:
            if isinstance(row, tuple):
                label, widget = row
                widget = widget.qobj
                if isinstance(label, Binder):
                    label = label.qobj
                qobj.addRow(label, widget)
            elif isinstance(row.qobj, QtGui.QWidget):
                qobj.addRow(row.qobj)

    def __repr__(self):
        args = ', '.join(map(repr, self.rows))
        if self.id:
            args += ', id={0.id!r}'.format(self)
        return '{0.__name__}({1})'.format(type(self), args)

    def _get_child_binders(self):
        children = []
        for row in self.rows:
            if isinstance(row, tuple):
                label, widget = row
                if isinstance(label, Binder):
                    children.append(label)
                children.append(widget)
            elif isinstance(row, Binder):
                children.append(row)
        return children


class WithLayout(Composite):
    """ A dumb `QWidget` wrapper with a child `Layout`.

    This is needed in some places where a true `QWidget` is needed instead of
    a `QLayout`.
    """
    qclass = QtGui.QWidget

    child_layout = Instance(Layout)

    def __init__(self, layout, **traits):
        super(WithLayout, self).__init__(child_layout=layout, **traits)

    def configure(self):
        super(WithLayout, self).configure()
        self.qobj.setLayout(self.child_layout.qobj)

    def __repr__(self):
        args = repr(self.child_layout)
        if self.id:
            args += ', id={0.id!r}'.format(self)
        return '{0.__name__}({1})'.format(type(self), args)


class ButtonGroup(Binder):
    """ A group of buttons.

    This is a special `Binder` used in the `button_groups=` keyword to `Bound`.
    `ButtonGroup` is not a widget, so it does not get put into the widget
    hierarchy. It is given the ID strings of the button `Binders` that belong
    to the group.
    """
    qclass = QtGui.QButtonGroup

    #: List of `Binder` ID strings or `(binder_id_str, qt_id_int)`
    button_ids = List(Either(Str, Tuple(Str, Int)))

    def __init__(self, *button_ids, **traits):
        button_ids = list(button_ids)
        super(ButtonGroup, self).__init__(button_ids=button_ids, **traits)

    def add_buttons_from_context(self, context):
        """ Pull out the required buttons from the context and add them.
        """
        for button_id in self.button_ids:
            qt_id = None
            if isinstance(button_id, tuple):
                button_id, qt_id = button_id
            button = context[button_id].qobj
            if qt_id is None:
                self.qobj.addButton(button)
            else:
                self.qobj.addButton(button, qt_id)


# These classes are not intended to be automatically looked up from their Qt
# classes.
_EXCLUDE_FROM_REGISTRY = [
    Composite,
    NChildren,
    SingleChild,
    SpanGridLayout,
    WithLayout,
]


for obj in list(vars().values()):
    if obj in _EXCLUDE_FROM_REGISTRY:
        continue
    if (isinstance(obj, type) and
            issubclass(obj, Binder) and
            obj is not Binder):
        binder_registry.push(obj.qclass, obj)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2014-2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------

import os
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))  # noqa

from traits.api import Any, Callable, HasTraits, NO_COMPARE, TraitError, \
    Unicode, on_trait_change
from traitsui.api import View

from qt_binder.api import Bound
from qt_binder.qt import QtGui
from qt_binder.raw_widgets import FormLayout, Label
from qt_binder.widgets import TextField, EnumDropDown


class PythonEvalidator(QtGui.QValidator):
    """ A QValidator that uses a Python evaluator function.

    If the function raises a TypeError, ValueError, or TraitError, the input
    will be considered unacceptable. Other errors will propagate.
    """

    def __init__(self, evaluate_func, *args, **kwds):
        self.evaluate_func = evaluate_func
        super(PythonEvalidator, self).__init__(*args, **kwds)

    def validate(self, text, pos):
        try:
            self.evaluate_func(text)
        except (TraitError, TypeError, ValueError):
            status = self.Intermediate
        else:
            status = self.Acceptable
        return status, text, pos


class Model(HasTraits):
    text = Unicode()
    # Make sure we notify every time this is set, even if ``120 == 120.0``.
    value = Any(comparison_mode=NO_COMPARE)
    evaluator_func = Callable(int)

    def default_traits_view(self):
        traits_view = View(
            Bound(
                FormLayout(
                    (u'Text:', TextField(id='field')),
                    (u'Kind:', EnumDropDown(
                        id='combo',
                        values=[
                            (int, u'Integer'),
                            (float, u'Float'),
                        ])),
                    (u'Value:', Label(id='value')),
                    (u'Text:', Label(id='text')),
                ),
                'field.value := object.text',
                'combo.value := object.evaluator_func',
                'field.validator << PythonEvalidator(object.evaluator_func)',
                'value.text << repr(object.value)',
                'text.text << object.text',
                extra_context=dict(PythonEvalidator=PythonEvalidator),
            ),
            resizable=True,
        )
        return traits_view

    @on_trait_change('text,evaluator_func')
    def _update_value(self):
        try:
            value = self.evaluator_func(self.text)
        except (TraitError, TypeError, ValueError):
            value = None
        self.value = value


def main():
    m = Model(text='120')
    m.configure_traits()

if __name__ == '__main__':
    main()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------

import os
import unittest

from pyface.ui.qt4.util.modal_dialog_tester import ModalDialogTester
from pyface.ui.qt4.util.gui_test_assistant import GuiTestAssistant
from traits.api import HasStrictTraits, Instance, Int, Unicode, \
    pop_exception_handler, push_exception_handler
from traitsui.api import View

from ..bound_editor import Bound
from ..qt import QtGui
from ..widgets import IntSlider, TextField, UIFile


def localfile(filename):
    return os.path.join(os.path.dirname(__file__), filename)


class TestUiLoader(unittest.TestCase, GuiTestAssistant):

    def setUp(self):
        super(TestUiLoader, self).setUp()
        push_exception_handler(reraise_exceptions=True)

    def tearDown(self):
        pop_exception_handler()
        super(TestUiLoader, self).tearDown()

    def test_load_ui_file(self):

        class _Model(HasStrictTraits):
            text = Unicode()
            number = Int()
            # Only for testing. Not a good idea for production, otherwise
            # multiple views will conflict with each other.
            ui_file = Instance(UIFile)

            def default_traits_view(self):
                traits_view = View(
                    Bound(
                        self.ui_file,
                        'lineEdit.value := object.text',
                        'widget.value := object.number',
                    ),
                )
                return traits_view

            def _ui_file_default(self):
                _ValidatedField = TextField(validator=QtGui.QIntValidator())
                _IntSlider = IntSlider(range=(0, 10))
                ui_file = UIFile(
                    localfile('form.ui'),
                    overrides={
                        'lineEdit': _ValidatedField,
                    },
                    insertions={
                        'widget': _IntSlider,
                    },
                )
                return ui_file

        m = _Model()
        tester = ModalDialogTester(lambda: m.edit_traits(kind='livemodal'))

        def _test(tester):
            with tester.capture_error():
                # Send valid and invalid data through the model to see if it
                # propagates to our overridden widget correctly.
                try:
                    w = tester.find_qt_widget(
                        type_=QtGui.QLineEdit,
                        test=lambda w: w.objectName() == u'lineEdit',
                    )
                    self.assertEqual(m.text, w.text())
                    m.text = u'10'
                    self.assertEqual(w.text(), u'10')
                    self.assertEqual(w.property('valid'), True)
                    m.text = u'abc'
                    self.assertEqual(w.text(), u'abc')
                    self.assertEqual(w.property('valid'), False)

                    w = tester.find_qt_widget(
                        type_=QtGui.QSlider,
                        test=lambda w: getattr(
                            w.parent(), 'objectName')() == u'widget',
                    )
                    self.assertEqual(m.number, w.value())
                    m.number = 5
                    self.assertEqual(w.value(), 5)
                    self.assertIsInstance(w, QtGui.QSlider)
                    self.assertIs(m.ui_file.widget.qobj, w)
                finally:
                    tester.close(accept=True)

        tester.open_and_run(when_opened=_test)


if __name__ == '__main__':
    unittest.main()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2014-2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------

import os
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))  # noqa

from traits.api import Any, HasTraits, NO_COMPARE
from traitsui.api import View

from qt_binder.api import Bound
from qt_binder.raw_widgets import ButtonGroup, FormLayout, GroupBox, Label, \
    PushButton, RadioButton, VBoxLayout


class Model(HasTraits):

    last_clicked_button = Any(comparison_mode=NO_COMPARE)
    last_pressed_button = Any(comparison_mode=NO_COMPARE)
    last_released_button = Any(comparison_mode=NO_COMPARE)

    def default_traits_view(self):
        traits_view = View(
            Bound(
                VBoxLayout(
                    GroupBox(
                        VBoxLayout(
                            PushButton(text=u'Push A', id='push_a'),
                            PushButton(text=u'Push B', id='push_b'),
                            PushButton(text=u'Push C', id='push_c'),
                        ),
                        title=u'Push Buttons',
                    ),
                    GroupBox(
                        VBoxLayout(
                            RadioButton(text=u'Radio A', id='radio_a',
                                        checked=True),
                            RadioButton(text=u'Radio B', id='radio_b'),
                            RadioButton(text=u'Radio C', id='radio_c'),
                        ),
                        title=u'Radio Buttons',
                    ),
                    GroupBox(
                        FormLayout(
                            (u'Clicked:', Label(id='clicked')),
                            (u'Pressed:', Label(id='pressed')),
                            (u'Released:', Label(id='released')),
                        ),
                        title=u'Events',
                    ),
                ),
                ('push_buttons.buttonClicked_QAbstractButton >> '
                 'object.last_clicked_button'),
                ('push_buttons.buttonPressed_QAbstractButton >> '
                 'object.last_pressed_button'),
                ('push_buttons.buttonReleased_QAbstractButton >> '
                 'object.last_released_button'),
                ('radio_buttons.buttonClicked_QAbstractButton >> '
                 'object.last_clicked_button'),
                ('radio_buttons.buttonPressed_QAbstractButton >> '
                 'object.last_pressed_button'),
                ('radio_buttons.buttonReleased_QAbstractButton >> '
                 'object.last_released_button'),
                ('clicked.text << (object.last_clicked_button).text() '
                 'if object.last_clicked_button else u""'),
                ('pressed.text << (object.last_pressed_button).text() '
                 'if object.last_pressed_button else u""'),
                ('released.text << (object.last_released_button).text() '
                 'if object.last_released_button else u""'),
                button_groups=dict(
                    push_buttons=ButtonGroup('push_a', 'push_b', 'push_c'),
                    radio_buttons=ButtonGroup('radio_a', 'radio_b', 'radio_c',
                                              exclusive=True),
                ),
            ),
            resizable=True,
            width=200,
            title=u'Button Groups',
        )
        return traits_view


def main():
    Model().configure_traits()

if __name__ == '__main__':
    main()
